<?php

namespace Drupal\content_synchronizer\Processors\Entity;

use Drupal\content_synchronizer\Events\ImportEvent;
use Drupal\content_synchronizer\Processors\ExportEntityWriter;
use Drupal\content_synchronizer\Processors\ExportProcessor;
use Drupal\content_synchronizer\Processors\ImportProcessor;
use Drupal\content_synchronizer\Service\EntityPublisher;
use Drupal\content_synchronizer\Service\GlobalReferenceManager;
use Drupal\Core\Plugin\PluginBase;
use Drupal\Core\Entity\Entity;
use Drupal\Core\Entity\ContentEntityType;
use Drupal\content_synchronizer\Processors\Type\TypeProcessorPluginManager;

/**
 * The entity processor base.
 */
class EntityProcessorBase extends PluginBase implements EntityProcessorInterface {

  const KEY_TRANSLATIONS = 'translations';

  protected $propertyIdsNotToExport = [
    'status',
    'revision',
    'revision_timestamp',
    'revision_uid',
    'revision_log',
    'revision_translation_affected',
    'created',
    // 'uuid',.
    'id',
  ];

  /**
   * The global reference manager service.
   *
   * @var \Drupal\content_synchronizer\Service\GlobalReferenceManager
   */
  protected $globalReferenceManager;

  /**
   * The type processor manager service.
   *
   * @var \Drupal\content_synchronizer\Processors\Type\TypeProcessorPluginManager
   */
  protected $typeProcessorManager;

  /**
   * The entity processor manager service.
   *
   * @var \Drupal\content_synchronizer\Processors\Entity\EntityProcessorPluginManager
   */
  protected $entityProcessorManager;

  /**
   * The entity publisher service.
   *
   * @var \Drupal\content_synchronizer\Service\EntityPublisher
   */
  protected $entityPublisher;

  /**
   * The current entity type.
   *
   * @var string
   */
  protected $entityType;

  /**
   * Export the entity and return the gid if exists, else  false.
   *
   * @param \Drupal\Core\Entity\Entity $entityToExport
   *   The entity to export.
   *
   * @return bool|string
   *   The gid if exported, False either
   */
  final public function export(Entity $entityToExport) {
    // If entity is exportable (content entity)
    if ($entityToExport->getEntityType() instanceof ContentEntityType) {
      // Get the entity gid.
      $gid = $this->getEntityGlobalReference($entityToExport);

      if (isset($entityToExport->contentSynchronizerIsExporting)) {
        return $gid;
      }
      else {
        $dataToExport = [];
        foreach ($this->getEntityTranslations($entityToExport) as $languageId => $translation) {
          // Tag the current entity has exporting in order to avoid circular dependencies.
          $translation->contentSynchronizerIsExporting = TRUE;
          $dataToExport[self::KEY_TRANSLATIONS][$languageId] = $this->getDataToExport($translation);
        }

        if (!empty($dataToExport)) {
          $entityToExport->contentSynchronizerGid = $dataToExport[ExportEntityWriter::FIELD_GID] = $gid;
          $dataToExport[ExportEntityWriter::FIELD_UUID] = $entityToExport->uuid();
          ExportProcessor::getCurrentExportProcessor()
            ->getWriter()
            ->write($entityToExport, $dataToExport);
          return $gid;
        }
      }
    }

    return FALSE;
  }

  /**
   * Return the data of the default language of the passed data.
   */
  public function getDefaultLanguageData(array $data) {
    if (count($data[self::KEY_TRANSLATIONS]) > 1) {
      foreach ($data[self::KEY_TRANSLATIONS] as $languageId => $translationData) {
        if (array_key_exists('default_langcode', $translationData) && $translationData['default_langcode'][0]['value'] == 1) {
          return $translationData;
        }
      }
    }

    // Get default data :
    $defaultData = reset($data[self::KEY_TRANSLATIONS]);

    // Filter on reference data field.
    $fieldDefinitions =
      \Drupal::entityTypeManager()
        ->getStorage($this->getGlobalReferenceManager()
          ->getEntityTypeFromGid($data[ExportEntityWriter::FIELD_GID]))
        ->getFieldStorageDefinitions();

    return array_intersect_key($defaultData, $fieldDefinitions);
  }

  /**
   * Return the entity translations.
   *
   * @param \Drupal\Core\Entity\Entity $entity
   *   The entity.
   *
   * @return array|\Drupal\Core\Language\LanguageInterface[]
   *   The array of translations.
   */
  protected function getEntityTranslations(Entity $entity) {
    $translations = [];
    foreach ($entity->getTranslationLanguages() as $languageId => $data) {
      $translations[$languageId] = \Drupal::service('entity.repository')
        ->getTranslationFromContext($entity, $languageId);
    }

    return $translations;
  }

  /**
   * Return a translation.
   *
   * @param string $languageId
   *   The language id.
   * @param \Drupal\Core\Entity\Entity $existingEntity
   *   The entity to translate.
   * @param array $dataToImport
   *   The data to import.
   */
  protected function createNewTranslation($languageId, Entity $existingEntity, array $dataToImport = []) {
    if ($existingEntity->isTranslatable()) {
      $translation = $existingEntity->addTranslation($languageId);
      $translation->uuid = \Drupal::service('uuid')->generate();
      return $translation;
    }
    return NULL;
  }

  /**
   * Create or update entity with data :.
   *
   * @param array $dataToImport
   *   The data to import.
   */
  final public function import(array $dataToImport) {
    $gid = $dataToImport[ExportEntityWriter::FIELD_GID];
    $uuid = $dataToImport[ExportEntityWriter::FIELD_UUID];

    // If the entity has already been imported then we don't have to do it again.
    $import = ImportProcessor::getCurrentImportProcessor()->getImport();
    if ($import->gidHasAlreadyBeenImported($gid)) {
      return $this->getGlobalReferenceManager()->getEntityByGid($gid);
    }

    // Tag as importing.
    ImportProcessor::getCurrentImportProcessor()
      ->getImport()
      ->tagHasImporting($gid);

    // Get the previous entity by gid.
    if ($existingEntity = $this->getGlobalReferenceManager()
      ->getExistingEntityByGidAndUuid($gid, $uuid)
    ) {
      if ($existingEntity) {
        $backup = clone($existingEntity);
      }
      /** @var \Drupal\Core\Entity\Entity $entity */
      if ($entity = $this->getEntityToImport($dataToImport, $existingEntity)) {
        $this->getEntityPublisher()->saveEntity($entity, $gid, $backup);
      }
    }
    else {
      /** @var \Drupal\Core\Entity\Entity $entity */
      if ($entity = $this->getEntityToImport($dataToImport, NULL)) {
        $this->getEntityPublisher()->saveEntity($entity, $gid, NULL);

        $this->getGlobalReferenceManager()
          ->createGlobalEntityByImportingEntityAndGid($entity, $gid);
      }
    }

    if ($entity) {
      // Tag as imported.
      ImportProcessor::getCurrentImportProcessor()
        ->getImport()
        ->tagHasImported($gid);

      $this->onEntityImported($gid, $entity);
    }

    return $entity;
  }

  /**
   * Callback when the entity has been imported.
   */
  protected function onEntityImported($gid, Entity $entity) {
    /** @var \Drupal\Component\EventDispatcher\ContainerAwareEventDispatcher $dispatcher */
    $dispatcher = \Drupal::service('event_dispatcher');

    $event = new ImportEvent();
    $event->setEntity($entity);
    $event->setGid($gid);

    $dispatcher->dispatch(ImportEvent::ON_ENTITY_IMPORTER, $event);
  }

  /**
   * Return the data to export.
   *
   * Get the array of data to export in array format :
   * [
   *    "property_1"=>[ "value1", "value2"]
   *    "property_2"=>[ "value1"]
   * ].
   *
   * @param \Drupal\Core\Entity\Entity $entityToExport
   *   The entity to export.
   *
   * @return array
   *   The data to export.
   */
  public function getDataToExport(Entity $entityToExport) {
    $dataToExport = [];

    // Get entity type keys :
    $contentEntityTypeKeys = $entityToExport->getEntityType()->getKeys();

    // Init properties not to export.
    $propertyIdsNotToExport = $this->getPropertiesIdsNotToExportList();
    $propertyIdsNotToExport += array_intersect_key($contentEntityTypeKeys, array_flip($propertyIdsNotToExport));

    // Init keys like bundles.
    /** @var \Drupal\Core\Entity\ContentEntityType $contentEntityType */
    foreach ($contentEntityTypeKeys as $key => $name) {
      if (!in_array($name, $propertyIdsNotToExport)) {
        if (method_exists($entityToExport, $key)) {
          $dataToExport[$name] = $entityToExport->$key();
        }
      }
    }

    foreach ($entityToExport->getTypedData()
      ->getProperties() as $propertyId => $propertyData) {
      // Check properties to export :
      if (!in_array($propertyId, $propertyIdsNotToExport)) {
        /** @var \Drupal\content_synchronizer\Processors\Type\TypeProcessorBase $plugin */
        if ($plugin = $this->getTypeProcessorManager()
          ->getInstanceByFieldType(get_class($propertyData))
        ) {
          if ($fieldDataToExport = $plugin->getExportedData($entityToExport->get($propertyId))) {
            $dataToExport[$propertyId] = $fieldDataToExport;
          }
        }
      }
    }

    return $dataToExport;
  }

  /**
   * Return the entity to import.
   *
   * @param array $data
   *   The data to import.
   * @param \Drupal\Core\Entity\Entity $entityToImport
   *   The existing entity to update.
   */
  public function getEntityToImport(array $data, Entity $entityToImport = NULL) {
    if ($entityToImport) {
      $backup = clone($entityToImport);
    }
    // Create Entity.
    if (is_null($entityToImport)) {
      try {
        $defaultData = $this->getDefaultLanguageData($data);

        $entityToImport = \Drupal::entityTypeManager()
          ->getStorage($this->getGlobalReferenceManager()
            ->getEntityTypeFromGid($data[ExportEntityWriter::FIELD_GID]))
          ->create($defaultData);
      }
      catch (\Exception $e) {
        drupal_set_message('Import Process : ' . $e->getMessage() . ' in "' . __METHOD__ . '()"', 'error');
        return NULL;
      }
    }

    // Properties not to import.
    $propertyIdsNotToImport = $this->getPropertiesIdsNotToExportList();

    // Get the existing translations.
    $alreadyExistingEntityTranslations = $this->getEntityTranslations($entityToImport);

    // Update data for each translation.
    foreach ($data[self::KEY_TRANSLATIONS] as $languageId => $translationData) {
      if (!array_key_exists($languageId, $alreadyExistingEntityTranslations)) {
        if ($translation = $this->createNewTranslation($languageId, $entityToImport, $translationData)) {
          $alreadyExistingEntityTranslations[$languageId] = $translation;
        }
        else {
          continue;
        }
      }

      $entityToUpdate = $alreadyExistingEntityTranslations[$languageId];

      // Parse each property of the entity.
      foreach ($entityToUpdate->getTypedData()
        ->getProperties() as $propertyId => $propertyData) {
        // Check properties to import :
        if (!in_array($propertyId, $propertyIdsNotToImport)) {

          /** @var \Drupal\content_synchronizer\Processors\Type\TypeProcessorBase $plugin */
          if ($plugin = $this->getTypeProcessorManager()
            ->getInstanceByFieldType(get_class($propertyData))
          ) {
            $plugin->initImportedEntity($entityToUpdate, $propertyId, $translationData);
          }
        }
      }

      // Save translation.
      if ($entityToImport->language()->getId() != $entityToUpdate->language()
        ->getId()
      ) {
        // $entityToUpdate->save();
        $this->getEntityPublisher()->saveEntity($entityToUpdate, NULL, $backup);
      }
    }

    return $entityToImport;
  }

  /**
   * Get the array of the property of the entity not to export.
   *
   * @return array
   *   The property not to export.
   */
  public function getPropertiesIdsNotToExportList() {
    return $this->propertyIdsNotToExport;
  }

  /**
   * Get the global reference entity.
   *
   * @param \Drupal\Core\Entity\Entity $entityToExport
   *   The entity to export.
   *
   * @return string
   *   The gid.
   */
  protected function getEntityGlobalReference(Entity $entityToExport) {
    $gid = $this->getGlobalReferenceManager()
      ->getEntityGlobalId($entityToExport);
    if (!$gid) {
      $gid = $this->getGlobalReferenceManager()
        ->createEntityGlobalId($entityToExport);
    }
    return $gid;
  }

  /**
   * Get the contentSyncManager.
   *
   * @return \Drupal\content_synchronizer\Service\GlobalReferenceManager
   *   The global reference manager service.
   */
  final protected function getGlobalReferenceManager() {
    if (!isset($this->globalReferenceManager)) {
      $this->globalReferenceManager = \Drupal::service(GlobalReferenceManager::SERVICE_NAME);
    }
    return $this->globalReferenceManager;
  }

  /**
   * Get the TypeProcessor plugin manager.
   *
   * @return \Drupal\content_synchronizer\Processors\Type\TypeProcessorPluginManager
   *   The type processor manager service.
   */
  protected function getTypeProcessorManager() {
    if (!$this->typeProcessorManager) {
      $this->typeProcessorManager = \Drupal::service(TypeProcessorPluginManager::SERVICE_NAME);
    }
    return $this->typeProcessorManager;
  }

  /**
   * Get the EntityProcessor plugin manager.
   *
   * @return \Drupal\content_synchronizer\Processors\Entity\EntityProcessorPluginManager
   *   The entity processor manager service.
   */
  protected function getEntityProcessorManager() {
    if (!$this->entityProcessorManager) {
      $this->entityProcessorManager = \Drupal::service(EntityProcessorPluginManager::SERVICE_NAME);
    }
    return $this->entityProcessorManager;
  }

  /**
   * Return the current entity type.
   */
  public function getEntityType() {
    return $this->entityType;
  }

  /**
   * Set the current entity type.
   */
  public function setEntityType($entityType) {
    $this->entityType = $entityType;
  }

  /**
   * Return the entity saver service.
   *
   * @return \Drupal\content_synchronizer\Service\EntityPublisher
   *   The Entity publisher service.
   */
  public function getEntityPublisher() {
    if (is_null($this->entityPublisher)) {
      $this->entityPublisher = \Drupal::service(EntityPublisher::SERVICE_NAME);
    }
    return $this->entityPublisher;
  }

}
